import base64 as b64
import inspect
import os
import platform
import random
import re
import shlex
import string
import subprocess as subproc
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

ALPHANUMERIC_CHARS = string.ascii_letters + string.digits
USERNAME_ASCII_CHARS = ALPHANUMERIC_CHARS + '_-'


def nlprint(*args, **kwargs) -> None:
    """Identical to print() but also print()'s before and after"""
    print()
    print(*args, **kwargs)
    print()


def is_jwt(s: Union[str, bytes]) -> bool:
    """Check if a string is a valid JWT (JSON Web Token).
    Args:
        s (Union[str, bytes]): Input suspected jwt bytes or string.
    Returns:
        bool: True if the input is a valid JWT, False otherwise.
    """
    # If the input is bytes, try to decode it as UTF-8
    if isinstance(s, bytes):
        try:
            s = s.decode('utf-8')
        except UnicodeDecodeError:
            return False

    parts = s.split('.')

    # A valid JWT should have exactly three parts separated by dots
    if len(parts) != 3:
        return False

    try:
        for part in parts:
            # Ensure that the part has proper padding to be a valid base64 string
            while len(part) % 4 != 0:
                part += '='
            # Attempt to decode the part as base64
            b64.urlsafe_b64decode(part)
    except ValueError:
        return False
    return True


def is_jwt_str(*args, **kwargs) -> bool:
    """legacy func passing to is_jwt"""
    return is_jwt(*args, **kwargs)


def ranstr(
    min_length: int = 16,
    max_length: Optional[int] = None,
    chars: Iterable = ALPHANUMERIC_CHARS,
    as_generator: bool = False,
) -> Union[Generator, str]:
    """Generates str with random chars between min and max length
    Args:
        min_length (int) = 16: The length of the string.
        max_length (Optional[int]): The maximum length of the string.
            If None, min_length is used with no variance on ranstr len
        chars (Iterable = ALPHANUMERIC_CHARS): Characters used for random str
        as_generator (bool) = False: Return a generator instead of a string
    Returns:
        Generator | str: The random str or generator for random str
    """
    if max_length is None:
        _length = min_length
    else:
        _length = random.randint(min_length, max_length)

    if as_generator:
        return (random.choice(chars) for _ in range(_length))
    return ''.join(random.choice(chars) for _ in range(_length))


def runcmd(
    cmd: str, output: bool = True, *args, **kwargs
) -> Optional[List[str]]:
    """Runs a single command in the shell.
    Args:
        cmd: escaped str that will be treated as if pasted into shell
        output (bool) = True:
            equivalent to check=True text=True capture_output=True
    Returns:
        List[str] | None: output of cmd if output=True, None otherwise
    """
    cmd = shlex.split(cmd)
    if output:
        return subproc.run(
            cmd, check=True, text=True, capture_output=True, *args, **kwargs
        ).stdout.splitlines()
    else:
        subproc.run(
            cmd, check=False, text=False, capture_output=False, *args, **kwargs
        )


def get_terminal_width(default: int = 80) -> int:
    """Gets current terminal width or default 80 if not detectable"""
    try:
        return os.get_terminal_size().columns
    except OSError:
        return default


def print_middle(
    obj: any, char: str = '=', noprint: bool = False, *args, **kwargs
) -> str:
    """prints object str in middle of line of $chars based on terminal width
    passes *args/**kwargs to print()
    Args:
        obj (Any): The object whose string representation needs to be printed.
        char (str): Char used for line default '='
        noprint (bool) = False: If True, don't print, just return str
    """
    terminal_width = get_terminal_width()
    _obj = str(obj)

    padding = (terminal_width - len(_obj) - 2) // 2

    # not using fstring for backwards compatibility
    result = '%s %s %s' % (char * padding, _obj, char * padding)

    if not noprint:
        print(result, *args, **kwargs)
    return result


def print_columns(
    iterable: Iterable[str],
    separator: str = "  ",
    terminal_width: Optional[int] = None,
) -> List[str]:
    """Print a list of objects in columns based on the terminal width.
    Args:
        iterable (Iterable): The iterable to be printed.
        separator (str) = "  ": The separator to be used between columns.
        terminal_width (Optional[int]): The terminal width to be used.
    Returns:
        List[str]: The list of strings that were printed.
    """
    termwidth = terminal_width or get_terminal_width()

    objs = [str(obj) for obj in iterable]
    longest = max(len(obj) for obj in objs)

    printed = []

    curline = ""
    while objs:
        curobj = objs.pop(0)
        curobj += " " * (longest - len(curobj))
        curobj += separator

        if len(curline) + len(curobj) > termwidth:
            print(curline)
            printed.append(curline)
            curline = ""
        else:
            curline += curobj

    return printed


def objinfo(obj: Any) -> None:
    """Print information about an object."""
    terminal_width = get_terminal_width()

    obj_type = type(obj)
    obj_name = obj.__name__ if hasattr(obj, "__name__") else "N/A"
    obj_attrs = sorted(
        str(attr) for attr in dir(obj) if not callable(getattr(obj, attr))
    )
    obj_methods = sorted(
        str(method) for method in dir(obj) if callable(getattr(obj, method))
    )
    obj_doc = inspect.getdoc(obj) or "No documentation available."
    obj_scope = (
        "Local"
        if obj in inspect.currentframe().f_back.f_locals.values()
        else "Global"
    )
    obj_size = sys.getsizeof(obj)
    obj_mutability = "Mutable" if hasattr(obj, "__dict__") else "Immutable"
    obj_identity = id(obj)

    header = "{} ({})".format(obj_name, obj_type)
    subheader = (
        "Size: {}".format(obj_size),
        "Scope: {}".format(obj_scope),
        "Mutability: {}".format(obj_mutability),
        "Identity: {}".format(obj_identity),
    )
    nlprint(print_middle(header, noprint=True))
    print_columns(subheader)

    # not using fstring for old versions of python
    nlprint(print_middle("Attributes", "-", noprint=True))
    print_columns(obj_attrs)
    nlprint(print_middle("Methods", "-", noprint=True))
    print_columns(obj_methods)
    nlprint(print_middle("Documentation", "-", noprint=True))
    print(obj_doc)
    nlprint('-' * terminal_width)


def default_repr(
    obj: Any, transform: Optional[Callable] = None, *args, **kwargs
) -> str:
    """Return a string representation of a custom Python object.
    This representation is constructed such that the object can be
    reconstructed from the returned string, ideally. Complex objects
    may not be able to be reconstructed.
    Args:
        obj (Any): The input Python object.
        transform (Optional[Callable]): A function that will return the repr
            string with modifications. *args/**kwargs will be passed to this func
    Returns:
        str: The string representation of the object.
    """
    # If the object has a __dict__ attribute, use that
    if hasattr(obj, '__dict__'):
        attributes = ', '.join(
            "{}={}".format(key, repr(value))
            for key, value in obj.__dict__.items()
            if not hasattr(value, '__call__') and not str(key).startswith("_")
        )
    # Otherwise, use dir() to gather potential attributes
    else:
        if isinstance(obj, int):
            return 'int(%s)' % obj
        elif isinstance(obj, float):
            return 'float(%s)' % obj
        elif isinstance(obj, str):
            return 'str(%s)' % obj
        elif isinstance(obj, set):
            return 'set(%s)' % obj
        elif isinstance(obj, (list, tuple, dict)):
            return str(obj)

        attributes = ', '.join(
            "{}={}".format(attr, getattr(obj, attr))
            for attr in dir(obj)
            if not hasattr(obj, '__call__') and not str(attr).startswith('_')
        )

    if transform is not None:
        return transform(obj, *args, **kwargs)
    return "{}({})".format(obj.__class__.__name__, attributes)


def typed_evar(name: str, default: Optional[Any] = None):
    """Return an environment variable with an assumed type. Type from
    the default value, if provided, will be prioritized, otherwise
    the type will be inferred in order of: bool, int, float, str.

    (CURDAY, 25) -> 25
    (CURDAY, 25.0) -> 25.0
    (CURDAY, '25.0') -> '25.0'
    (CURDAY, None) -> '25'

    """
    varval = os.environ.get(name)
    if varval is None:
        return default

    # use default's type
    if default is not None:
        vartype = type(default)

        # bool gets special treatment
        if vartype is bool:
            if varval.lower() in ('1', 'true'):
                return True
            elif varval.lower() in ('0', 'false', '-1'):
                return False
            else:
                raise ValueError(
                    "Invalid boolean value for environment variable %s: %s"
                    % (name, varval)
                )
        try:
            return vartype(varval)
        except:
            pass

    # otherwise assume type using a few simple types
    if varval.lower() in ('true', 'false'):
        return varval.lower() == 'true'

    for vartype in (int, float):
        try:
            return vartype(varval)
        except ValueError:
            continue

    return varval



from .main import (
    ALPHANUMERIC_CHARS,
    USERNAME_ASCII_CHARS,
    nlprint,
    ranstr,
    print_middle,
    get_terminal_width,
    print_columns,
    objinfo,
    default_repr,
    typed_evar,
    is_jwt,
    is_jwt_str,
    runcmd,
)



import os
import sys
import unittest
import random as ran
from decimal import Decimal
from os.path import abspath, dirname
from subprocess import CompletedProcess
from typing import Generator
from unittest.mock import MagicMock, call, patch

sys.path.insert(0, dirname(dirname(abspath(__file__))))

from myfuncs import (
    ALPHANUMERIC_CHARS,
    default_repr,
    get_terminal_width,
    is_jwt_str,
    is_jwt,
    nlprint,
    objinfo,
    print_columns,
    print_middle,
    ranstr,
    runcmd,
    typed_evar,
)

_valid_jwtstr = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ0ZXN0IjoidGVzdCJ9.MZZ7UbJRJH9hFRdBUQHpMjU4TK4XRrYP5UxcAkEHvxE'


class TestRunCmd(unittest.TestCase):
    def test_runcmd_with_output(self):
        with patch(
            'subprocess.run',
            return_value=CompletedProcess(
                args=['echo', 'Hello, World!'],
                returncode=0,
                stdout='Hello, World!',
                stderr='',
            ),
        ):
            result = runcmd('echo Hello, World!', output=True)
        self.assertEqual(result, ['Hello, World!'])

    def test_runcmd_without_output(self):
        with patch('subprocess.run'):
            result = runcmd('echo Hello, World!', output=False)

        self.assertIsNone(result)

    def test_runcmd_with_special_characters(self):
        special_cmd = "echo \"Hello\" && touch /tmp/evil.txt"
        with patch('subprocess.run') as mock_run:
            runcmd(special_cmd, output=False)

        mock_run.assert_called_once_with(
            ['echo', 'Hello', '&&', 'touch', '/tmp/evil.txt'],
            check=False,
            text=False,
            capture_output=False,
        )


class TestPrintMiddle(unittest.TestCase):
    @patch("myfuncs.main.get_terminal_width", return_value=50)
    def test_print_middle_no_print(self, *args):
        tstr = '012345678901'
        clen = (50 - len(f' {tstr} ')) // 2
        with patch('builtins.print') as mock_print:
            result = print_middle(tstr, noprint=True)
        self.assertEqual(result, f'{clen * "="} {tstr} {clen * "="}')
        mock_print.assert_not_called()


class TestGetTerminalWidth(unittest.TestCase):
    def test_default_terminal_width(self):
        with patch('os.get_terminal_size', side_effect=OSError):
            result = get_terminal_width()
        self.assertEqual(result, 80)

    def test_actual_terminal_width(self):
        with patch(
            'os.get_terminal_size', return_value=os.terminal_size((100, 24))
        ):
            result = get_terminal_width()
        self.assertEqual(result, 100)


class TestPrintColumns(unittest.TestCase):
    teststrs = ['a' * i for i in range(1, 20)]

    def print_columns(self):
        result = print_columns(self.teststrs, terminal_width=80)
        ran.shuffle(self.teststrs)
        ranresult = print_columns(self.teststrs, terminal_width=80)

        for r in [result, ranresult]:
            self.assertEqual(len(r), 4)
            self.assertTrue(max(len(s) for s in r) == min(len(s) for s in r))


class TestRanStr(unittest.TestCase):
    def test_string_length(self):
        result = ranstr(5, 10)
        self.assertTrue(5 <= len(result) <= 10)

    def test_string_characters(self):
        chars = 'abcd'
        result = ranstr(5, 10, chars=chars)
        for char in result:
            self.assertIn(char, chars)

    def test_return_generator(self):
        result = ranstr(5, 10, as_generator=True)
        self.assertIsInstance(result, Generator)
        generated_string = ''.join(result)
        self.assertTrue(5 <= len(generated_string) <= 10)

    def test_return_string(self):
        result = ranstr(5, 10)
        self.assertIsInstance(result, str)
        self.assertTrue(5 <= len(result) <= 10)


class TestCustomReprFunction(unittest.TestCase):
    class MyClass:
        def __init__(self, a, b):
            self.a = a
            self.b = b

        def method(self):
            pass

    class AnotherClass:
        c = "class attribute"

        def __init__(self, x):
            self.x = x

    def test_simple_class(self):
        instance = self.MyClass(1, "test")
        representation = default_repr(instance)
        expected_repr = "MyClass(a=1, b='test')"
        self.assertEqual(representation, expected_repr)

    def test_class_with_class_attribute(self):
        instance = self.AnotherClass(5)
        representation = default_repr(instance)
        expected_repr = "AnotherClass(x=5)"
        self.assertEqual(representation, expected_repr)

    def test_builtin_type_without_dict(self):
        value = 123
        representation = default_repr(value)
        expected_repr = "int(123)"
        self.assertEqual(representation, expected_repr)

    def test_list(self):
        lst = [1, 2, 3]
        representation = default_repr(lst)
        expected_repr = "[1, 2, 3]"
        self.assertEqual(representation, expected_repr)

    def test_set(self):
        st = {1, 2, 3}
        representation = default_repr(st)
        expected_repr = "set({1, 2, 3})"
        self.assertEqual(representation, expected_repr)

    def test_float(self):
        value = 123.45
        representation = default_repr(value)
        expected_repr = "float(123.45)"
        self.assertEqual(representation, expected_repr)

    def test_tuple(self):
        tpl = (1, 2, 3)
        representation = default_repr(tpl)
        expected_repr = "(1, 2, 3)"
        self.assertEqual(representation, expected_repr)

    def test_dict(self):
        d = {'a': 1, 'b': 2}
        representation = default_repr(d)
        expected_repr = "{'a': 1, 'b': 2}"
        self.assertEqual(representation, expected_repr)

    def test_with_private_attribute(self):
        class WithPrivate:
            def __init__(self, x):
                self._x = x

        instance = WithPrivate(5)
        representation = default_repr(instance)
        expected_repr = "WithPrivate()"
        self.assertEqual(representation, expected_repr)


TVAR = 'TEST_VAR'


class TestTypedEvar(unittest.TestCase):
    def setUp(self):
        if TVAR in os.environ:
            del os.environ[TVAR]

    def test_no_default_and_no_value(self):
        self.assertIsNone(typed_evar(TVAR))

    def test_no_default(self):
        os.environ[TVAR] = 'Hello World'
        self.assertEqual(typed_evar(TVAR), 'Hello World')

    def test_no_default_int(self):
        os.environ[TVAR] = '123'
        self.assertEqual(typed_evar(TVAR), 123)

    def test_no_default_float(self):
        os.environ[TVAR] = '123.456'
        self.assertEqual(typed_evar(TVAR), 123.456)

    def test_with_default_and_no_value(self):
        self.assertEqual(typed_evar(TVAR, default=42), 42)

    def test_default_bool_with_true_value(self):
        os.environ[TVAR] = 'true'
        self.assertEqual(typed_evar(TVAR, default=False), True)

    def test_default_bool_with_false_value(self):
        os.environ[TVAR] = 'false'
        self.assertEqual(typed_evar(TVAR, default=True), False)

    def test_default_bool_with_invalid_value(self):
        os.environ[TVAR] = 'invalid_bool'
        with self.assertRaises(ValueError):
            typed_evar(TVAR, default=False)

    def test_infer_bool_true(self):
        os.environ[TVAR] = 'true'
        self.assertEqual(typed_evar(TVAR), True)

    def test_infer_bool_false(self):
        os.environ[TVAR] = 'false'
        self.assertEqual(typed_evar(TVAR), False)

    def test_default_bool_with_1_value(self):
        os.environ[TVAR] = '1'
        self.assertEqual(typed_evar(TVAR, default=False), True)

    def test_default_bool_with_0_value(self):
        os.environ[TVAR] = '0'
        self.assertEqual(typed_evar(TVAR, default=True), False)

    def test_infer_int(self):
        os.environ[TVAR] = '123'
        self.assertEqual(typed_evar(TVAR), 123)

    def test_infer_float(self):
        os.environ[TVAR] = '123.456'
        self.assertEqual(typed_evar(TVAR), 123.456)

    def test_infer_str(self):
        os.environ[TVAR] = 'Hello World'
        self.assertEqual(typed_evar(TVAR), 'Hello World')

    def test_default_decimal_with_value(self):
        os.environ[TVAR] = '123.456'
        self.assertEqual(
            typed_evar(TVAR, default=Decimal('0.0')), Decimal('123.456')
        )

    def test_infer_bool_as_int_1(self):
        os.environ[TVAR] = '1'
        self.assertEqual(typed_evar(TVAR, default=True), True)

    def test_infer_bool_as_int_0(self):
        os.environ[TVAR] = '0'
        self.assertEqual(typed_evar(TVAR, default=True), False)

    def test_infer_bool_as_int_neg1(self):
        os.environ[TVAR] = '-1'
        self.assertEqual(typed_evar(TVAR, default=True), False)

    def test_is_jwt_str_legacy(self):
        self.assertTrue(is_jwt_str(_valid_jwtstr))


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(
    name='myfuncs',
    version='1.5.3',
    packages=find_packages(),
    install_requires=[],
    author='Cary Carter',
    author_email='ccarterdev@gmail.com',
    description=(
        'Personal utility functions that I use across different codebases.'
    ),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/cc-d/myfuncs',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)
